#include "pam_ru.h"

/* Necessary for become to work properly with pam_ru */
struct ruauthdata {
  char *ph_path;
  char *ru_agent;
};

int pam_sm_authenticate(pamh, flags, argc, argv)
	pam_handle_t 	*pamh;
	int 	flags;
	int 	argc;
	const char 	**argv;
{
   struct pam_conv *conv;
   struct pam_message *msgvec[1];
   struct pam_response *respvec[1];
   struct pam_message msg;
   int ret, i, top = 0;
   char *tmp;
   char *user;
   char *password;
   char *spassword;
   char *nl;

   unsigned long time_mul = 15000000;
   int debdev = 0;
   int errno;

   msgvec[0] = &msg;

   for (i = 0; i < argc; i++) {
      if (!strcmp("debug", argv[i])) debdev = 1;
      else if (!strcmp("expire", argv[i])) time_mul = 60;
      else if (!strncmp("exp=", argv[i], 4)) {
         tmp = strdup(argv[i]);
         tmp = tmp + 4;
         if (tmp) {
            long time_arg = atol(tmp);
            if (time_arg <= 0) {
               time_arg = 60;
            }
            else if (time_arg > 15000000) {
               time_arg = 15000000;
            }
            time_mul = (unsigned long)time_arg;
         }
         else time_mul = 60;
      }
   }

   if (debdev) {
      for (i = 0; i < argc; i++) debout("ru_save - arg %s", argv[i]);
      debout("ru_save - current expiration time: %ld minutes", time_mul);
   }

/*
 * Retrieving the necessary password information
 */

   if (debdev) debout("ru_save - Getting conversation.");

#ifdef sun
   if (pam_get_item(pamh, PAM_CONV, (void **)&conv) != PAM_SUCCESS) {
#else
   if (pam_get_item(pamh, PAM_CONV, (const void **)&conv) != PAM_SUCCESS) {
#endif
      return PAM_SERVICE_ERR;
   }

   if (debdev) debout("ru_save - Getting user.");

#ifdef sun
   if (pam_get_user(pamh, (char **)&user, NULL) != PAM_SUCCESS || !user || !*user) {
#else
   if (pam_get_user(pamh, (const char **)&user, NULL) != PAM_SUCCESS || !user || !*user) {
#endif 
      return PAM_USER_UNKNOWN;
   }

   if (debdev) {
      debout("ru_save - user = %s", user);
      debout("ru_save - Getting password.");
   }

#ifdef sun
   if (pam_get_item(pamh, PAM_AUTHTOK, (void **)&password) != PAM_SUCCESS || !password) {
#else
   if (pam_get_item(pamh, PAM_AUTHTOK, (const void **)&password) != PAM_SUCCESS || !password) {
#endif

      if (debdev) debout("ru_save - password wasn't set.");
      
      top = 1;

      msg.msg_style = PAM_PROMPT_ECHO_OFF;
      msg.msg = "Token: ";

#ifdef sun
      if ((*(conv->conv)) (1, (struct pam_message **)msgvec, respvec, conv->appdata_ptr) != PAM_SUCCESS) {
#else
      if ((*(conv->conv)) (1, (const struct pam_message **)msgvec, respvec, conv->appdata_ptr) != PAM_SUCCESS) {
#endif
         return PAM_AUTH_ERR;
      }
     
      password = respvec[0]->resp;

      // strip newline (if any) from password
      if (nl = strchr(password, '\n')) {
         *nl = '\0';
      }
   }

   if (password) {

      if (debdev) {
         debout("ru_save - password has been set.");
         debout("ru_save - It is now %s", password);
      }

      if (!strncmp(password, "save ", 5)) {

         if (debdev) debout("ru_save - save pass = %s", password);
         
	 if (pam_set_item(pamh, PAM_OLDAUTHTOK, (void *)password) != PAM_SUCCESS) {
            return PAM_SERVICE_ERR;
         }

         spassword = strdup(password + 5);

         if (debdev) {
           debout("ru_save - %s", spassword);
	   debout("ru_save - calling ru_save.");
           debout("ru_save - user = %s, password = %s, old_pass = %s", user, spassword, password);
	 }

         ret = ru_save(user, password, pamh, 0, debdev, time_mul);
         if (debdev) debout("ru_save - return code for ru_save = %d", ret);
         
	 if (!ret) {
            
            spassword = NULL;
            if (pam_set_item(pamh, PAM_OLDAUTHTOK, (void *)spassword) != PAM_SUCCESS) {
               return PAM_SERVICE_ERR;
            }

            return PAM_SUCCESS;
         }
         else {
            if (pam_set_item(pamh, PAM_AUTHTOK, (void *)spassword) != PAM_SUCCESS) {
               return PAM_SERVICE_ERR;
	    }

 	    if (debdev) debout("ru_save - set_item PAM_AUTHTOK = %s", spassword);
            
	    return PAM_AUTH_ERR;
         }
      }
      else {

         if (debdev) {
            debout("ru_save - not a save password.  moving on.");
            debout("ru_save - user = %s", user);
            debout("ru_save - pass = %s", password);
         }

         if (top && (pam_set_item(pamh, PAM_AUTHTOK, (void *)password) != PAM_SUCCESS)) {
            return PAM_SERVICE_ERR;
         }

         return PAM_AUTH_ERR;
      }

   }

   else return PAM_AUTH_ERR;

}

int pam_sm_setcred(pamh, flags, argc, argv)
	pam_handle_t 	*pamh;
	int 	flags;
	int 	argc;
	const char 	**argv;
{
   return PAM_SUCCESS;
}
