#include <sys/time.h>

#include "pam_ru.h"

/* Necessary for become to work properly with pam_ru */
struct ruauthdata {
  char *ph_path;
  char *ru_agent;
};

int errno;

int ru_save(char *user, char *password, pam_handle_t *pamh, int save, 
            int debdev, unsigned long time_mul) {
   
   int cachefile; 
   char passbuf[80] = { (char)NULL };
   char tmp[300];
   char *service;
   char *rhost;
   struct stat statb;
   struct timeval timeb;
   int become = 0;

   char *group_h;
   char *user_h;

   if (gettimeofday(&timeb, 0)) {
      if (debdev) debout("ru_save - error getting the time of day.");
      return -1;
   } 

   if (stat("/var/spool/save-cache/", &statb)) {
      return -1;
   }

   if (debdev) debout("ru_save - Inside pam_save");

   /* Get the service */
   if (pam_get_item(pamh, PAM_SERVICE, (void *)&service) == PAM_SUCCESS) {
      
      snprintf(tmp, 300, "/var/spool/save-cache/%s-cache", service);
      
      if (stat(tmp, &statb)) mkdir(tmp, 0700);

      /* Get the remote host */
#ifdef sun
      if (pam_get_item(pamh, PAM_RHOST, (void **)&rhost) == PAM_SUCCESS) {
#else
      if (pam_get_item(pamh, PAM_RHOST, (const void **)&rhost) == PAM_SUCCESS) {
#endif

         /* Get IP address info given the host name */
         if (rhost) {
            
            struct addrinfo *result;
            int returncode;
            char ipbuf[INET_ADDRSTRLEN];

            returncode = getaddrinfo(rhost, NULL, NULL, &result);

            if (returncode || !result) {
               debout("ru_save - problem getting ip address info.");
               return -1;
            }
            
            /* Get the human-readable numeric ip address */
            returncode = getnameinfo(result->ai_addr, result->ai_addrlen, ipbuf, INET_ADDRSTRLEN, NULL, 0, NI_NUMERICHOST); 
            if (returncode) {
               debout("ru_save - problem getting ip address string.");
               return -1;
            }            

            /*  Decide whether the account is a become account or a normal account.
             *  The file name will be of the form <user>-<ip address> or 
             *  <become group>-<user>-<IP address>.
             */
            if (pam_get_data(pamh, "become_user", (const void **)&user_h) == PAM_SUCCESS) {
         
               become = 1;

               if (debdev) debout("ru_save - working with a become account.");
                  
               if (pam_get_data(pamh, "become_group", (const void **)&group_h) == PAM_SUCCESS) {
                   
                  if (debdev)  debout("ru_save - user handle %s", user_h);
            
                  snprintf(tmp, 300, "/var/spool/save-cache/%s-cache/%s-%s-%s", service, group_h, user_h, ipbuf);
                   
                  if (debdev) debout("ru_save - %s", tmp);
               }
            }
            else {
               snprintf(tmp, 300, "/var/spool/save-cache/%s-cache/%s-%s", service, user, ipbuf);
               if (debdev) debout("ru_save - %s", tmp);
            }

            freeaddrinfo(result);

            /* To save or not to save */
            if (!save) {

               /* Get the status of the cached password file */
               if (!stat(tmp, &statb)) {

                  /* Check that the password hasn't expired */
                  if (timeb.tv_sec < (statb.st_mtime + time_mul * 60)) {
                  
                     cachefile = open(tmp, O_RDONLY);
     
                     /* Check that the file exists */
                     if (cachefile) {
                     
                        /* Read the cached password */
                        if (read (cachefile, passbuf, sizeof(passbuf)) == sizeof(passbuf) 
                              && !strncmp(passbuf, password, sizeof(passbuf))) {
                              
                           close(cachefile);
                              
                           if (debdev) debout("ru_save - cached password ok from %s", user);

                           if (become) {

                              if (!group_h || (pam_set_item(pamh, PAM_USER, group_h) != PAM_SUCCESS)) {
                                 debout("ru_save - reset of PAM_USER to become group failed");
                                 return -1;
                              }

                              if (debdev) debout("ru_save - set_item PAM_USER = %s", group_h);
                           }

                           return 0;
                           
                        }
                        else {
                           if (debdev) debout("ru_save - incorrect password from %s", user);
                           return -1;
                        } 
                     }
                     else {
                        if (debdev) debout("ru_save - could not open the file for %s", user);
                        return -1;
                     }
                  }
                  else {
                     debout("ru_save - expired password for %s", user);
                     if (debdev) {
                        debout("ru_save - current time: %ld", timeb.tv_sec);
                        debout("ru_save - expiration time: %ld", 
                              statb.st_mtime + time_mul * 60);
                     }
                     return -1;
                  }
               }
               else {
                  debout("ru_save - Cannot find %s's cached password", user);
                  if (debdev) {
                     debout("ru_save - If you are using an application that does not");
                     debout("ru_save - run as root (apache), check to see if the");
                     debout("ru_save - directory permissions are set properly.");
                  }
                  return -1;
               }
            }

            /* Save the password */
            else {
               if (password) {

                  cachefile = open(tmp, O_WRONLY|O_CREAT, 0600);

                  if (cachefile) {
               
                     snprintf(passbuf, 80, "%s", password);

                     if (write(cachefile, (const void *)&passbuf,sizeof(passbuf)) == sizeof(passbuf)) {
                        if (debdev) {
                           debout("ru_save - caching password for %s", user);
                        }
                     }
                     else {
                        close(cachefile);
                        debout("ru_save - Could not write to save file.");
                        debout("ru_save - Reason: %s", strerror(errno));
                        return -1;
                     }

                     close(cachefile);
                     return 0;

                  }
                  else {
                     debout("ru_save - could not create cache file.");
                     return -1;
                  }
               }
               else {
                  /* Shouldn't happen */
                  debout("ru_save - Invalid password."); 
                  return -1;
               }
            }   
         }
         else {
            /* Shouldn't happen */
            debout("ru_save - The value returned by PAM_RHOST is invalid");
            return -1;
         }
      }
      else {
         debout("ru_save - PAM_RHOST is not set.");
         return -1;
      }
   }
   else {
      debout("ru_save - Could not find the service name.");
      return -1;
   }
}
